import std/[nre, strtabs, strutils]
import ./notation
import zero_functional

type
  ChordChartMetadata* = object
    title*: string
    key*: Key
    optionalProps: StringTableRef

  ChordChart* = ref object
    rawContent*: string
    metadata*: ChordChartMetadata
    nodes*: seq[ChordChartNode]

  ChordChartNodeKind* = enum
    ccnkSection,
    ccnkLine,
    ccnkWord,
    ccnkNote,
    ccnkColBreak,
    ccnkPageBreak,
    ccnkTransposeKey,
    ccnkRedefineKey,
    ccnkNone

  ChordChartChord* = object
    original*: Option[string]
    root*: ScaleDegree
    flavor*: Option[string]
    bass*: Option[ScaleDegree]

  ChordChartNode* = ref object
    case kind: ChordChartNodeKind
    of ccnkSection:
      sectionName*: string
      remainingSectionLine*: Option[string]
      sectionContents*: seq[ChordChartNode]
    of ccnkLine:
      line*: seq[ChordChartNode]
    of ccnkWord:
      spaceBefore*: bool
      spaceAfter*: bool
      chord*: Option[ChordChartChord]
      word*: Option[string]
    of ccnkNote:
      note*: string
      inclInLyrics*: bool
    of ccnkColBreak: discard
    of ccnkPageBreak: discard
    of ccnkTransposeKey:
      transposeSteps*: int
    of ccnkRedefineKey:
      newKey*: Key
    of ccnkNone: discard

  ParserContext = ref object
    lines: seq[string]
    curKey: Key
    curLineNum: int
    curSection: ChordChartNode
    unparsedLineParts: seq[string]

let EMPTY_CHORD_CHART_NODE* = ChordChartNode(kind: ccnkNone)

func `[]`*(ccmd: ChordChartMetadata, key: string): string =
  ccmd.optionalProps[key]

func contains*(ccmd: ChordChartMetadata, key: string): bool =
  return key == "title" or key == "key" or ccmd.optionalProps.contains(key)

iterator pairs*(ccmd: ChordChartMetadata): tuple[key, value: string] =
  yield ("title", ccmd.title)
  yield ("key", $ccmd.key)
  for p in ccmd.optionalProps.pairs: yield p

func kind*(ccn: ChordChartNode): ChordChartNodeKind = ccn.kind

func dump*(m: ChordChartMetadata, indent = ""): string =
  return indent & "Metadata\p" & join(m.pairs --> map(indent & "  " & it.key & ": " & it.value), "\p")

func dump*(ccn: ChordChartNode, indent = ""): string =
  case ccn.kind

  of ccnkSection:
    let formattedChildren = ccn.sectionContents --> map(dump(it, indent & "  "))
    result = indent & "Section -- " & ccn.sectionName & "\p" &
      formattedChildren.join("\p")

  of ccnkColBreak: result = indent & "Column Break"
  of ccnkPageBreak: result = indent & "Page Break"

  of ccnkTransposeKey:
    result = indent & "Transpose key by " & $ccn.transposeSteps

  of ccnkRedefineKey:
    result = indent & "Redefine key to " & $ccn.newKey

  of ccnkLine:
    result = indent & "Line  "
    for child in ccn.line:
      let formattedChild = dump(child, indent)
      if child.kind == ccnkWord: result &= formattedChild
      else: result &= formattedChild & "\p"

  of ccnkWord:
    result = ""
    if ccn.spaceBefore: result &= " "

    if ccn.chord.isSome:
      let chord = ccn.chord.get
      result &= "["
      if chord.original.isSome and chord.flavor.isNone and
         chord.bass.isNone:
        result &= chord.original.get
      else:
        result &= $chord.root
        if chord.flavor.isSome:
          result &= "_" & chord.flavor.get
        if chord.bass.isSome:
          result &= "/" & $chord.bass.get
      result &= "]"

    if ccn.word.isSome: result &= ccn.word.get
    if ccn.spaceAfter: result &= " "

  of ccnkNote:
    result = indent & "Note "
    if not ccn.inclInLyrics: result &= "(chords only) "
    result &= ccn.note

  of ccnkNone:
    result &= indent & "NONE"

func `$`*(cc: ChordChart): string =
  result = "ChordChart\p" &
    dump(cc.metadata, "  ") & "\p" &
    join(cc.nodes --> map(dump(it, "  ")), "\p")

# -----------------------------------------------------------------------------
#                                 PARSER
# -----------------------------------------------------------------------------

func hasNextLine(ctx: ParserContext): bool =
  ctx.unparsedLineParts.len > 1 or
  ctx.curLineNum + 1 < ctx.lines.len

func nextLine(ctx: var ParserContext): string =
  if ctx.unparsedLineParts.len > 0:
    result = ctx.unparsedLineParts[0]
    ctx.unparsedLineParts.delete(0)
  else:
    ctx.curLineNum += 1
    result = ctx.lines[ctx.curLineNum]

func pushPartialsToParse(ctx: var ParserContext, parts: varargs[string]): void =
  ctx.unparsedLineParts &= parts

template makeError(ctx: ParserContext, msg: string): untyped =
  newException(ValueError,
    "error parsing input at line " & $ctx.curLineNum & ":\p\t" & msg)

template addToCurSection(n: ChordChartNode): untyped =
  if ctx.curSection.kind == ccnkSection: ctx.curSection.sectionContents.add(n)
  else: result.add(n)

# see https://regexr.com/8f4ru
let CHORD_REGEX =
  "([b#♭♮𝄫𝄪]?[A-G1-7][b#♭♮𝄫𝄪]?)" &               #      chord root
  "((min|maj|aug|dim|sus|6\\/9|[mM1-9#b♭♮𝄫𝄪Δ+oøoø°𝆩][0-9]?|\\([1-9#b♭]+\\))*)" &  # chord flavor/type
  "(\\/([b#♭♮𝄫𝄪]?[A-G1-7][b#♭♮𝄫𝄪]?))?"             # optional bass
let CHORD_PAT = re(CHORD_REGEX)

proc parseChord*(
    ctx: ParserContext,
    chordValue: string
  ): Option[ChordChartChord] =

  let m = chordValue.match(CHORD_PAT)
  if m.isNone: none[ChordChartChord]()
  else:
    let flavor =
      if m.get.captures.contains(1) and m.get.captures[1].len > 0:
        some(m.get.captures[1])
      else: none[string]()

    let bass =
      if m.get.captures.contains(4) and m.get.captures[4].len > 0:
        some(toScaleDegree(ctx.curKey, parseSpelledPitch(m.get.captures[4])))
      else: none[ScaleDegree]()

    return some(ChordChartChord(
      original: some(chordValue),
      root: toScaleDegree(ctx.curKey, parseSpelledPitch(m.get.captures[0])),
      flavor: flavor,
      bass: bass))

let METADATA_LINE_PAT = re"^([^:]+):(.*)$"
let METADATA_END_PAT = re"^-+$"
proc parseMetadata(ctx: var ParserContext): ChordChartMetadata =

  var title = "MISSING"
  var optProps = newStringTable()
  var songKey: Option[Key]

  while ctx.curLineNum < ctx.lines.len:
    let line = ctx.nextLine

    if line.match(METADATA_END_PAT).isSome:
      break

    let m = line.match(METADATA_LINE_PAT)
    if m.isNone:
      raise ctx.makeError("expected metadata property or ending marker")

    let key = m.get.captures[0].strip.tolower
    let value = m.get.captures[1].strip
    if key == "title": title = value
    elif key == "key":
      let parts = value.split(" ", 1)
      songKey = some(Key(
        tonic: parseSpelledPitch(parts[0]),
        mode:
          if parts.len > 1: parseMode(parts[1])
          else: Ionian))
      if songKey.isNone:
        raise ctx.makeError("unrecognized key: " & value)
    else: optProps[key] = value

  if title == "MISSING":
    raise ctx.makeError("metadata is missing the 'title' property")

  if songKey.isNone:
    raise ctx.makeError("metadata is missing the 'key' property")

  result = ChordChartMetadata(
    title: title,
    key: songKey.get,
    optionalProps: optProps)

const KNOWN_SECTION_NAMES = [
  "chorus", "verse", "bridge", "breakdown", "vamp", "intstrumental",
  "interlude", "intro", "outtro", "ending", "end", "tag", "prechorus",
  "pre-chorus", "pre chorus"
]

let SECTION_LINE_PAT = re(
  "^((" & "((?i)" &                               # case insensitive
  KNOWN_SECTION_NAMES.join("|") & ")" &           # known names
  "|[[:upper:]]{3,}" &                            # all upper-case words
  ") ?\\d*)" &                                    # numeric suffix (Verse 2)

  # Allow notes or other text to follow on the same line
  "(.*)$"
)
let COL_BREAK_PAT = re"\s*COLUMN_BREAK\s*$"
let PAGE_BREAK_PAT = re"\s*PAGE_BREAK\s*$"
let TRANSPOSE_PAT = re"\s*TRANSPOSE KEY ([+-]\d+)\s*$"
let REDEFINE_KEY_PAT = re"\s*REDEFINE KEY ([+-]\d+)\s*$"
#let NOTE_PAT = re"^(.*)({{[^}]+}}|{[^}]+})(.*)$"
let NOTE_START_PAT = re"\{\{"
let NOTE_END_PAT = re"\}\}"
let SPACE_PAT = re"\s"
let CHORD_IN_LYRICS_PAT = re"(\w+)(\[.+)"
let CHORD_AND_LYRICS_PAT = re"^\[([^\]]*)\]([^\s\[]+)(.*)$"
let BRACED_CHORD_PAT = re"^\[([^\]]+)\]$"
let NAKED_CHORDS_ONLY_PAT = re("^\\s*(" & CHORD_REGEX & "\\s*\\|*\\s*)+$")

proc readNote(ctx: var ParserContext, firstLine: string): tuple[pre, note, post: string] =
  let startLineNum = ctx.curLineNum

  result = ("", "", "")

  let startMatch = firstLine.find(NOTE_START_PAT)
  var endMatch = firstLine.find(NOTE_END_PAT)

  if startMatch.isNone: return

  if startMatch.get.matchBounds.a > 0:
    result.pre = firstLine[0..<startMatch.get.matchBounds.a]

  if endMatch.isSome:
    result.note = firstLine[(startMatch.get.matchBounds.b + 1)..<endMatch.get.matchBounds.a]
    result.post = firstLine[(endMatch.get.matchBounds.b + 1)..^1]
  else:
    # if we don't find the end of the note, then we need to read more lines
    # until we do
    result.note = firstLine[(startMatch.get.matchBounds.b + 1)..^1]

    while ctx.lines.len > ctx.curLineNum and ctx.hasNextLine:
      let line = ctx.nextLine
      endMatch = line.find(NOTE_END_PAT)
      if endMatch.isSome:
        result.note &= line[0..<endMatch.get.matchBounds.a]
        result.post = line[(endMatch.get.matchBounds.b + 1)..^1]
        return
      else: result.note &= line

    raise ctx.makeError("a note section was started on line " &
      $startLineNum & " and never closed")

proc parseLineParts(ctx: ParserContext, parts: varargs[string]): seq[ChordChartNode] =
  result = @[]
  for p in parts:
    var m = p.match(SPACE_PAT)
    if m.isSome:
      result &= ctx.parseLineParts(p.splitWhitespace)
      continue

    m = p.match(CHORD_IN_LYRICS_PAT)
    if m.isSome:
      result &= ctx.parseLineParts(m.get.captures[0], m.get.captures[1])
      continue

    m = p.match(CHORD_AND_LYRICS_PAT)
    if m.isSome:
      result.add(ChordChartNode(
        kind: ccnkWord,
        spaceAfter: true, #FIXME
        spaceBefore: true, #FIXME
        chord: ctx.parseChord(m.get.captures[0]),
        word: some(m.get.captures[1])))
      result &= ctx.parseLineParts(m.get.captures[2])
      continue

    m = p.match(BRACED_CHORD_PAT)
    if m.isSome:
      result.add(ChordChartNode(
        kind: ccnkWord,
        spaceAfter: false, #FIXME
        spaceBefore: false, #FIXME
        chord: ctx.parseChord(m.get.captures[0]),
        word: none[string]()))
      continue

    if p.len > 0:
      result.add(ChordChartNode(
        kind: ccnkWord,
          spaceAfter: true, #FIXME
          spaceBefore: true, #FIXME
          chord: none[ChordChartChord](),
          word: some(p)))

proc parseLine(lentCtx: var ParserContext, line: string): ChordChartNode =
  let ctx = lentCtx
  result = ChordChartNode(kind: ccnkLine)

  var m = line.match(NAKED_CHORDS_ONLY_PAT)
  if m.isSome:
    result.line = line.splitWhitespace --> map(
      ChordChartNode(
        kind: ccnkWord,
        spaceAfter: false, #FIXME
        spaceBefore: false, #FIXME
        chord: ctx.parseChord(it.strip),
        word: none[string]()))
    return

  m = line.match(NOTE_START_PAT)
  if m.isSome:
    let (pre, note, post) = lentCtx.readNote(line)
    result.line = ctx.parseLineParts(pre) &
      @[ChordChartNode(
        kind: ccnkNote,
        inclInLyrics: true,
        note: note)] &
      ctx.parseLineParts(post)
    return

  else: result.line = ctx.parseLineParts(line.splitWhitespace)


proc parseBody(ctx: var ParserContext): seq[ChordChartNode] =
  result = @[]
  while ctx.hasNextLine:
    var line = ctx.nextLine

    var m = line.match(TRANSPOSE_PAT)
    if m.isSome:
      addToCurSection(ChordChartNode(
        kind: ccnkTransposeKey,
        transposeSteps: parseInt(m.get.captures[0])))
      continue

    m = line.match(REDEFINE_KEY_PAT)
    if m.isSome:
      addToCurSection(ChordChartNode(
        kind: ccnkRedefineKey,
        newKey: ctx.curKey.transpose(parseInt(m.get.captures[0]))))
      continue

    m = line.match(COL_BREAK_PAT)
    if m.isSome:
      result.add(ChordChartNode(kind: ccnkColBreak))
      continue

    m = line.match(PAGE_BREAK_PAT)
    if m.isSome:
      result.add(ChordChartNode(kind: ccnkPageBreak))
      ctx.curSection = EMPTY_CHORD_CHART_NODE
      continue

    m = line.match(SECTION_LINE_PAT)
    if m.isSome:
      let captures = m.get.captures.toSeq
      ctx.curSection = ChordChartNode(
        kind: ccnkSection,
        sectionName:
          if captures[0].isSome: captures[0].get.strip
          else: raise ctx.makeError("unknown error parsing section header: " & line),
        remainingSectionLine:
          if captures[3].isSome: some(captures[3].get.strip)
          else: none[string](),
        sectionContents: @[])
      result.add(ctx.curSection)
      continue

    else:
      addToCurSection(ctx.parseLine(line))
      continue


proc parseChordChart*(s: string): ChordChart =
  var parserCtx = ParserContext(
    lines: s.splitLines,
    curLineNum: -1,
    curSection: EMPTY_CHORD_CHART_NODE,
    unparsedLineParts: @[])

  let metadata = parseMetadata(parserCtx)
  parserCtx.curKey = metadata.key

  result = ChordChart(
    rawContent: s,
    metadata: metadata,
    nodes: parseBody(parserCtx))
